-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library IEEE;
use ieee.std_logic_1164.all;			
use ieee.std_logic_unsigned.all; 
use ieee.std_logic_arith.all;
use work.sha3_blake_package.all;
use work.sha3_pkg.all;

entity blake_fsm1 is 
	generic ( BS : integer := 512 );
	port (
	clk : in std_logic;
	rst : in std_logic;		
	
	-- datapath sigs
	zcRemainder, zc1, zc0, final_segment	: in std_logic;
    extra_block : in std_logic;

    stbot : out std_logic;
	ein, lc, ec : out std_logic;
	eth, dt, etl  : out std_logic;		 
    load_id : out std_logic;
	
	-- Control communication	  
	load_next_block	: in std_logic;
	block_ready_set : out std_logic;
	msg_end_set : out std_logic;  
	
	-- FIFO communication
	src_ready : in std_logic;
	src_read    : out std_logic	
	);
end blake_fsm1;

architecture counter of blake_fsm1 is
	
	constant MW				: integer := BS;		-- message width	
	constant MWSEG			: integer := MW/64; 
	constant LOG2MWSEG 		: integer := log2( MWSEG ); 
	constant LOG2MWSEGZEROS	: std_logic_vector(LOG2MWSEG-1 downto 0) := (others => '0');
	
	-- counter
	signal zjfin, lj, ej : std_logic;		 
	signal wc : std_logic_vector(LOG2MWSEG-1 downto 0);

	-- fsm sigs
	type state_type is ( reset, wait_for_mid, wait_for_header1, load_block, wait_for_load1, wait_for_header2, wait_for_load2 );
	signal cstate_fsm1, nstate : state_type; 	  
		
	signal clr_f, set_f, f : std_logic;	
	
	signal set_empty_block, clr_empty_block, empty_block : std_logic;	 
	signal set_stbot, clr_stbot : std_logic;	 

	signal last_msg_block, set_last_msg_block, clr_last_msg_block : std_logic;
begin		
		
	sr_final_block : sr_reg 
	port map ( rst => rst, clk => clk, set => set_f, clr => clr_f, output => f);	

	-- fsm1 counter		
	word_counter_gen : countern generic map ( n => log2mwseg ) port map ( clk => clk, rst => '0', load => lj, en => ej, input => LOG2MWSEGZEROS, output => wc);
	zjfin <= '1' when wc = conv_std_logic_vector(mwseg-1,log2mwseg) else '0';
	
	-- state process
	cstate_proc : process ( clk )
	begin
		if rising_edge( clk ) then 
			if rst = '1' then
				cstate_fsm1 <= reset;
			else
				cstate_fsm1 <= nstate;
			end if;	   
		end if;
	end process;
	
	nstate_proc : process ( cstate_fsm1, src_ready, load_next_block, zjfin, zc0, final_segment, f, empty_block, zcRemainder, last_msg_block  )
	begin
		case cstate_fsm1 is
			when reset =>
				nstate <= wait_for_mid;
			when wait_for_mid =>
				if ( src_ready = '0' ) then
					nstate <= wait_for_header1;
				else
					nstate <= wait_for_mid;
				end if;
			when wait_for_header1 =>
				if ( src_ready = '1' ) then
					nstate <= wait_for_header1;
				elsif ( final_segment = '1' ) then
					nstate <= wait_for_header2;
				else 
					nstate <= wait_for_load1;
				end if;
			when wait_for_header2 =>
				if (src_ready = '1') then
					nstate <= wait_for_header2;
				else
					nstate <= wait_for_load1;
				end if;	 
			when wait_for_load1 =>
				if (load_next_block = '0') then
					nstate <= wait_for_load1;
				else
					nstate <= load_block;
				end if;	 				
			when load_block =>		
				if ((src_ready = '1') or (src_ready = '0' and zjfin = '0') ) then
					nstate <= load_block;
				elsif (src_ready = '0' and zjfin = '1' and zc0 = '1' and f = '1' and empty_block = '0' and (zcRemainder = '0' or last_msg_block = '1')) then
					nstate <= wait_for_load2;																
				elsif (src_ready = '0' and zjfin = '1' and f = '0' and zc0 = '1') then
					nstate <= wait_for_header1;										
				else
					nstate <= wait_for_load1;
				end if;									
		   when wait_for_load2 =>
		   		 if (load_next_block = '0') then
					nstate <= wait_for_load2;
				else
					nstate <= wait_for_mid;
				end if;
		end case;
	end process;
	
	-- fsm output			 
	sr_empty_block : sr_reg 
		port map ( rst => rst, clk => clk, set => set_empty_block, clr => clr_empty_block, output => empty_block);	
	set_empty_block <= '1' when (cstate_fsm1 = wait_for_load1 and ((extra_block = '1' and zc0 = '1') or (zc1 = '1' and zcRemainder = '0')) and load_next_block = '1' and empty_block = '0' and f = '1') else '0';
	clr_empty_block <= '1' when (cstate_fsm1 = reset) or (cstate_fsm1 = wait_for_load1 and load_next_block = '1' and empty_block = '1') else '0';
	
	sr_last_block : sr_reg 
		port map ( rst => rst, clk => clk, set => set_last_msg_block, clr => clr_last_msg_block, output => last_msg_block);	
    set_last_msg_block <= '1' when (cstate_fsm1 = wait_for_load1 and zc0 = '1' and zcRemainder = '1' and load_next_block = '1') else '0';

	clr_last_msg_block <= clr_stbot;
    load_id <= '1' when (cstate_fsm1 = wait_for_mid and src_ready = '0') else '0';
			
	dt <= '1' when ((cstate_fsm1 = reset) or 					
					(cstate_fsm1 = wait_for_load1 and empty_block = '1' and load_next_block = '1') or
					(cstate_fsm1 = wait_for_load2 and load_next_block = '1')) else '0';					
 		
	src_read <= '1' when 	((cstate_fsm1 = wait_for_mid and src_ready = '0') or 
							(cstate_fsm1 = wait_for_header1 and src_ready = '0') or 
							(cstate_fsm1 = wait_for_header2 and src_ready = '0') or 
							(cstate_fsm1 = wait_for_load1 and load_next_block = '1' and src_ready = '0' ) or							
							(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '0') or 
							(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '1')) else '0';
		
	ein <= '1' when ((cstate_fsm1 = wait_for_load1 and load_next_block = '1' and src_ready = '0' ) or
				 	(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '0') or 
					(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '1')) else '0';
						 
	ej <= '1' when 	((cstate_fsm1 = wait_for_load1 and load_next_block = '1' and src_ready = '0' ) or
				 	(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '0') or 
					(cstate_fsm1 = load_block and src_ready = '0' and zjfin = '1')) else '0';
						 
    block_ready_set <= '1' when ((cstate_fsm1 = load_block and src_ready = '0' and zjfin = '1')) else '0';
						
	msg_end_set <= '1' when (cstate_fsm1 = wait_for_load2 and load_next_block = '1' and f = '1' )  else '0';
	
	clr_f <= '1' when (cstate_fsm1 = wait_for_load2 and load_next_block = '1')  else '0';	   		
	set_f <= '1' when (cstate_fsm1 = wait_for_header2) else '0';

	lc <= '1' when (cstate_fsm1 = wait_for_header1 or cstate_fsm1 = wait_for_header2) else '0';
		
	lj <= '1' when ((cstate_fsm1 = reset)) else '0';  
		
	
	ec <= '1' when 	(cstate_fsm1 = wait_for_load1 and zc0 = '0' and load_next_block = '1') else '0';		
	eth <= '1' when (cstate_fsm1 = wait_for_load1 and zc0 = '0' and load_next_block = '1') else '0';		
	etl <= '1' when (cstate_fsm1 = wait_for_header2) else '0';	
	
	sr_stbot : sr_reg 
		port map ( rst => rst, clk => clk, set => set_stbot, clr => clr_stbot, output => stbot);	
	set_stbot <= '1' when (cstate_fsm1 = wait_for_load1 and f = '1' and zc0 = '1' and empty_block = '0' and load_next_block = '1') else '0';				  
	clr_stbot <= '1' when (cstate_fsm1 = wait_for_header1) else '0';
end counter;