-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;	   
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity blake_fsm2 is 
	generic (
		ppl : integer := 4;
		version : integer := SHA3_ROUND3;  
		h : integer := 256;
		FF : integer := 2
	);
	port (
		-- global
		clk : in std_logic;
		rst : in std_logic;
		
		-- datapath	  
		load_hash, eh : out std_logic_vector(ppl-1 downto 0);
		round : out std_logic_vector(3+FF-1 downto 0);		
		load_h : out std_logic; 
		lm : out std_logic;
		sf : out std_logic;
		smsg : out std_logic_vector(log2(ppl)-1 downto 0);	  
		sperm : out std_logic;		-- Select permute =)
		-- control				   
			--fsm1 hand shake signals
		block_ready_clr : out std_logic_vector(ppl-1 downto 0);		
		msg_end_clr 	: out std_logic_vector(ppl-1 downto 0);
		
		block_ready		: in std_logic_vector(ppl-1 downto 0);
		msg_end 		: in std_logic_vector(ppl-1 downto 0);	
				
			--fsm3 handshake signals
		output_busy_set  : out std_logic_vector(ppl-1 downto 0);
		output_busy		 : in  std_logic_vector(ppl-1 downto 0)
	);							   
end blake_fsm2;

architecture beh of blake_fsm2 is			  
	constant roundnr 			: integer := get_roundnr( h, version )*(FF);
	constant log2roundnr		: integer := 3+FF+1;
	constant log2roundnrzeros : std_logic_vector(log2roundnr-1 downto 0) := (others => '0') ;
		
	signal pc, pc_pre : std_logic_vector(log2roundnr-1 downto 0);
	signal ziroundnr,ziroundnr_pre,zi0, zi15, ei, li : std_logic;
	
	signal output_data_s : std_logic_vector(ppl-1 downto 0);	
	signal block_ready_true, valid_set, valid_clr, valid, valid_reg : std_logic_vector(ppl-1 downto 0);
	signal sf_clr, sf_set, sf_sig : std_logic_vector(ppl-1 downto 0);
	signal ppl_stage_cnt : std_logic_vector(log2(ppl)-1 downto 0);
	
begin
	sperm <= pc_pre(0);
	-- Internals
	proc_counter_gen : countern generic map ( N => log2roundnr ) port map ( clk => clk, rst => rst, load => li, en => ei, input => log2roundnrzeros, output => pc_pre);
	
	pplstage_cnt: process ( clk )
	begin
		if rising_edge( clk ) then		
			if rst = '1' or ei = '1' or li = '1' then
				ppl_stage_cnt <= (others => '0');
			else
				ppl_stage_cnt <= ppl_stage_cnt + 1;
			end if;
			ziroundnr <= ziroundnr_pre;
			pc <= pc_pre;
		end if;
	end process;

	smsg <= ppl_stage_cnt;
	
	ziroundnr_pre <= '1' when pc_pre = conv_std_logic_vector(roundnr,log2roundnr) else '0';
	zi0 <= '1' when pc = conv_std_logic_vector(0,log2roundnr) else '0';
	zi15 <= '1' when pc = conv_std_logic_vector(8,log2roundnr) else '0';
	round <= pc_pre(3+FF-1 downto 0);

	lm <= li;
	li <= '1' when (ziroundnr_pre = '1' and ppl_stage_cnt = 0) else '0';
	ei <= '1' when (ppl_stage_cnt = ppl-1) else '0';
	
	gg1: for i in 0 to ppl-1 generate	 				
		sf_set(i) <= '1' when (zi15 = '1' and msg_end(i)  = '1')  else '0';		
		
		sf_gen : sr_reg 
			generic map ( init => '1' )
			port map ( rst => rst, clk => clk, set => sf_set(i), clr => sf_clr(i), output => sf_sig(i));		
		
		valid_gen : sr_reg 
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => valid_set(i), clr => valid_clr(i), output => valid(i));		
	end generate;		
	
	ppl2 : if ppl = 2 generate
		sf <= sf_sig(0) when ppl_stage_cnt(0) = '0' else sf_sig(1);
	end generate; 
	ppl4: if ppl = 4 generate
		with ppl_stage_cnt(1 downto 0) select
		sf <= 	sf_sig(0) when "00",
				sf_sig(1) when "01",
				sf_sig(2) when "10",
				sf_sig(3) when others;
	end generate;
			
	load_h <= '1' when ((ziroundnr = '1') or (zi0 = '1' and ppl_stage_cnt /= conv_std_logic_vector(0,log2(ppl)))) else '0';
	
	
	eh <= sf_clr;
	
	-- Valid and SF	   		
	output_busy_set <= output_data_s;
	load_hash		<= output_data_s;	   
	msg_end_clr		<= output_data_s;
	
	output_data_s(0) <= '1' when (ziroundnr = '1' and msg_end(0) = '1' and output_busy(0) = '0' and valid_reg(0) = '1') else '0';	
	valid_set(0) <= '1' when (ziroundnr = '1' and ppl_stage_cnt = "00" and block_ready_true(0) = '1') else '0';	
	valid_clr(0) <= '1' when (li = '1' and block_ready_true(0) = '0') else '0';
	ctrlsig: for i in 1 to ppl-1 generate
		output_data_s(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and msg_end(i) = '1' and output_busy(i) = '0' and valid_reg(i) = '1' ) else '0';
		valid_set(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '1') else '0';
		valid_clr(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '0') else '0';
	end generate;	
	
	vreg : process ( clk ) 
	begin
		if rising_edge( clk ) then
			if li = '1' then  
				block_ready_true <= block_ready;
				valid_reg <= valid;
			end if;
		end if;
	end process;
	block_ready_clr <= valid_set;
	sf_clr <= valid_set;
end beh;																						  

			  